# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# NOTE(markwash): These functions are used in the base tests cases to
# set up the db api implementation under test. Rather than accessing them
# directly, test modules should use the load and reset functions below.
get_db = None
reset_db = None


def load(get_db_fn, reset_db_fn):
    global get_db, reset_db
    get_db = get_db_fn
    reset_db = reset_db_fn


def reset():
    global get_db, reset_db
    get_db = None
    reset_db = None
